
import java.applet.Applet;
import java.applet.AppletContext;
import java.awt.*;
import java.net.MalformedURLException;
import java.net.URL;

public class ARQSimulation extends Applet
    implements Runnable
{

    public void init()
    {
        currentAnimationType = new AnimationType();
        setLayout(new BorderLayout());
        setBackground(Color.lightGray);
        frameNum = new SynCounter();
        canvas = new DrawingArea(currentAnimationType, this);
        FramedPanel framedpanel = new FramedPanel(canvas);
        add("Center", framedpanel);
        Panel panel = new Panel();
        panel.setLayout(new GridLayout(5, 1));
        Panel panel1 = new Panel();
        Panel panel2 = new Panel();
        Panel panel3 = new Panel();
        Panel panel4 = new Panel();
        Panel panel5 = new Panel();
        idleRQbtn = new Button("Idle RQ");
        contRQbtn = new Button("Cont. RQ (No Errors)");
        SELREPlostPKTbtn = new Button("Selective Repeat (Lost Packet)");
        SELREPlostACKbtn = new Button("Selective Repeat (Lost ACK)");
        idleRQlostPktbtn = new Button("Idle RQ (Lost PKT)");
        idleRQlostAckbtn = new Button("Idle RQ (Lost ACK)");
        helpbtn = new Button("Help");
        freezebtn = new Button("Freeze Simulation");
        freezebtn.disable();
        resetbtn = new Button("Reset");
        resetbtn.disable();
        startbtn = new Button("Restart Simulation");
        startbtn.disable();
        bandwidthtxtfd = new TextField("50", 5);
        frametxtfd = new TextField("1024", 8);
        delaytxtfd = new TextField("50", 5);
        panel1.add(idleRQbtn);
        panel1.add(idleRQlostAckbtn);
        panel1.add(idleRQlostPktbtn);
        panel2.add(contRQbtn);
        panel2.add(SELREPlostPKTbtn);
        panel2.add(SELREPlostACKbtn);
        panel3.add(freezebtn);
        panel3.add(startbtn);
        panel3.add(resetbtn);
        panel3.add(helpbtn);
        panel4.add(new Label("Bandwidth (10 - 100 Mb/sec)"));
        panel4.add(new Label("Frame size (64 - 4500 bytes)"));
        panel4.add(new Label("One Way Delay (0.5 - 500 msec)"));
        panel5.add(bandwidthtxtfd);
        panel5.add(new Label("                       "));
        panel5.add(frametxtfd);
        panel5.add(new Label("                        "));
        panel5.add(delaytxtfd);
        panel.add(panel1);
        panel.add(panel2);
        panel.add(panel3);
        panel.add(panel4);
        panel.add(panel5);
        add("South", panel);
        validate();
    }

    public void start()
    {
        if(!frozen && currentAnimatorThread == null)
        {
            currentAnimatorThread = new Thread(this);
            currentAnimatorThread.setPriority(1);
            currentAnimatorThread.start();
        }
    }

    public void stop()
    {
        currentAnimatorThread = null;
    }

    public void run()
    {
        long l = System.currentTimeMillis();
        for(; Thread.currentThread() == currentAnimatorThread; frameNum.incrementCounter())
        {
            canvas.repaint();
            try
            {
                l += threadDelay;
                Thread.sleep(Math.max(0L, l - System.currentTimeMillis()));
            }
            catch(InterruptedException _ex) { }
        }

    }

    public boolean action(Event event, Object obj)
    {
        if(event.target instanceof Button)
        {
            if(event.target == idleRQbtn)
            {
                currentAnimationType.setType(1);
                startSim();
                return true;
            }
            if(event.target == idleRQlostAckbtn)
            {
                currentAnimationType.setType(7);
                startSim();
                return true;
            }
            if(event.target == idleRQlostPktbtn)
            {
                currentAnimationType.setType(8);
                startSim();
                return true;
            }
            if(event.target == contRQbtn)
            {
                currentAnimationType.setType(2);
                startSim();
                return true;
            }
            if(event.target == SELREPlostPKTbtn)
            {
                currentAnimationType.setType(3);
                startSim();
                return true;
            }
            if(event.target == SELREPlostACKbtn)
            {
                currentAnimationType.setType(4);
                startSim();
                return true;
            }
            if(event.target == freezebtn)
            {
                freezebtn.disable();
                startbtn.enable();
                frozen = true;
                stop();
                return true;
            }
            if(event.target == startbtn)
            {
                startbtn.disable();
                freezebtn.enable();
                frozen = false;
                start();
                return true;
            }
            if(event.target == resetbtn)
            {
                resetSim();
                return true;
            }
            if(event.target == helpbtn)
            {
                URL url = getDocumentBase();
                URL url1 = null;
                boolean flag = true;
                try
                {
                    url1 = new URL(url, "ARQhelp.html");
                }
                catch(MalformedURLException _ex)
                {
                    flag = false;
                    String as[] = {
                        " Unable to form Help URL.", " Help CAN NOT be directly accessed from applet.", " The help file, ARQSimHelp.html, can be viewed", " using a Web Browser. ARQSimHelp.html is in the", " same directory as this Applet."
                    };
                    java.awt.Frame frame = FindFrame.getFrame(this);
                    AlertWindow alertwindow = new AlertWindow(frame, "Warning!", as);
                    alertwindow.move(250, 200);
                    alertwindow.pack();
                    alertwindow.show();
                }
                if(flag)
                {
                    AppletContext appletcontext = getAppletContext();
                    appletcontext.showDocument(url1, "Help");
                }
                return true;
            } else
            {
                return false;
            }
        } else
        {
            return false;
        }
    }

    private void startSim()
    {
        idleRQbtn.disable();
        idleRQlostPktbtn.disable();
        idleRQlostAckbtn.disable();
        contRQbtn.disable();
        SELREPlostPKTbtn.disable();
        SELREPlostACKbtn.disable();
        bandwidthtxtfd.disable();
        frametxtfd.disable();
        delaytxtfd.disable();
        freezebtn.enable();
        resetbtn.enable();
        frozen = false;
        String s = bandwidthtxtfd.getText();
        byte byte0 = 100;
        byte byte1 = 10;
        try
        {
            bandwidth = Integer.parseInt(s);
        }
        catch(NumberFormatException _ex)
        {
            bandwidth = 50;
            bandwidthtxtfd.setText("50");
        }
        if(bandwidth < byte1 || bandwidth > byte0)
        {
            bandwidth = 50;
            bandwidthtxtfd.setText("50");
        }
        s = frametxtfd.getText();
        byte byte2 = 64;
        char c = '\u1194';
        try
        {
            framesize = Integer.parseInt(s);
        }
        catch(NumberFormatException _ex)
        {
            framesize = 1024;
            frametxtfd.setText("1024");
        }
        if(framesize < byte2 || framesize > c)
        {
            framesize = 1024;
            frametxtfd.setText("1024");
        }
        s = delaytxtfd.getText();
        float f = 500F;
        float f1 = 0.5F;
        try
        {
            framedelay = (new Float(s)).floatValue();
        }
        catch(NumberFormatException _ex)
        {
            framedelay = 50F;
            delaytxtfd.setText("50");
        }
        if(framedelay < f1 || framedelay > f)
        {
            framedelay = 50F;
            delaytxtfd.setText("50");
        }
        float f2 = 250F;
        threadDelay = 50 + (int)((framedelay / f) * f2);
        sim = new SimARQ(frameNum, getBackground(), currentAnimationType, bandwidth, framesize, framedelay);
        start();
    }

    protected void resetSim()
    {
        if(currentAnimatorThread != null)
            currentAnimatorThread.stop();
        currentAnimatorThread = null;
        sim.deallocate();
        sim = null;
        resetbtn.disable();
        startbtn.disable();
        freezebtn.disable();
        idleRQbtn.enable();
        idleRQlostPktbtn.enable();
        idleRQlostAckbtn.enable();
        SELREPlostACKbtn.enable();
        contRQbtn.enable();
        SELREPlostPKTbtn.enable();
        bandwidthtxtfd.enable();
        frametxtfd.enable();
        delaytxtfd.enable();
        frozen = true;
        currentAnimationType.setType(0);
        frameNum.resetCounter();
        canvas.repaint();
        System.gc();
    }

    public ARQSimulation()
    {
        frozen = true;
    }

    private boolean frozen;
    private Thread currentAnimatorThread;
    private int threadDelay;
    private AnimationType currentAnimationType;
    private int bandwidth;
    private int framesize;
    private float framedelay;
    protected SimARQ sim;
    protected SynCounter frameNum;
    private DrawingArea canvas;
    private Button idleRQbtn;
    private Button contRQbtn;
    private Button SELREPlostPKTbtn;
    private Button SELREPlostACKbtn;
    private Button idleRQlostPktbtn;
    private Button idleRQlostAckbtn;
    private Button freezebtn;
    private Button startbtn;
    private Button resetbtn;
    private Button helpbtn;
    private TextField bandwidthtxtfd;
    private TextField frametxtfd;
    private TextField delaytxtfd;
}

class AlertWindow extends Dialog
{

    public AlertWindow(Frame frame, String as[])
    {
        this(frame, "", as);
    }

    public AlertWindow(Frame frame, String s, String as[])
    {
        super(frame, s, true);
        setBackground(Color.white);
        setResizable(false);
        setLayout(new GridLayout(as.length + 1, 1));
        for(int i = 0; i < as.length; i++)
            add(new Label(as[i]));

        Panel panel = new Panel();
        panel.add(new Button("OK"));
        add(panel);
    }

    public boolean action(Event event, Object obj)
    {
        if(event.target instanceof Button)
        {
            if("OK".equals(obj))
            {
                dispose();
                return true;
            } else
            {
                return false;
            }
        } else
        {
            return false;
        }
    }

    public boolean handleEvent(Event event)
    {
        if(event.id == 201)
        {
            dispose();
            return true;
        } else
        {
            return super.handleEvent(event);
        }
    }
}



class AnimationType
{

    public int getType()
    {
        return currtype;
    }

    public void setType(int i)
    {
        if(i >= 0 && i <= 8)
        {
            currtype = i;
            return;
        } else
        {
            currtype = 0;
            return;
        }
    }

    AnimationType()
    {
    }

    public static final int NONE = 0;
    public static final int IDLERQ = 1;
    public static final int CONTRQ = 2;
    public static final int SELREPLOSTPKT = 3;
    public static final int SELREPLOSTACK = 4;
    public static final int GOBACKNLOSTACK = 5;
    public static final int GOBACKNLOSTPKT = 6;
    public static final int IDLERQLOSTACK = 7;
    public static final int IDLERQLOSTPKT = 8;
    private int currtype;
}


class DrawingArea extends Canvas
{

    public DrawingArea(AnimationType animationtype, ARQSimulation arqsimulation)
    {
        currAnimation = animationtype;
        appletContainer = arqsimulation;
    }

    public void update(Graphics g)
    {
        paint(g);
    }

    public void paint(Graphics g)
    {
        Dimension dimension = size();
        int i1 = (int)(0.02D * (double)dimension.width);
        byte byte0 = 15;
        if(isCanvasSizeOK(g, dimension))
        {
            g.setColor(Color.black);
            g.setFont(new Font("TimesRoman", 0, 12));
            FontMetrics fontmetrics = g.getFontMetrics();
            int i = fontmetrics.getHeight();
            int j = fontmetrics.getDescent();
            int k = fontmetrics.getMaxAdvance();
            int l = fontmetrics.getAscent();
            g.drawString("Packet", i1, byte0);
            g.drawString("Number", i1, byte0 + i);
            g.drawString("______", i1, byte0 + i + j / 2);
            g.drawString("1", i1 + k, byte0 + 2 * i);
            g.drawString("2", i1 + k, byte0 + 3 * i);
            g.drawString("3", i1 + k, byte0 + 4 * i);
            g.drawString("4", i1 + k, byte0 + 5 * i);
            g.drawString("Timer", i1 + 6 * k, byte0);
            g.drawLine(i1 + 4 * k, byte0 + 2 * j, i1 + 10 * k, byte0 + 2 * j);
            g.drawString("|", i1 + 4 * k, byte0 + 3 * j);
            g.drawString("|", i1 + 10 * k, byte0 + 3 * j);
            g.drawString("SENDER", (int)(0.5D * (double)dimension.width) - 2 * k, (int)(0.45000000000000001D * (double)dimension.height));
            g.drawString("RECEIVER", (int)(0.5D * (double)dimension.width) - 2 * k, (int)(0.97999999999999998D * (double)dimension.height));
            g.drawString("TIME ", (int)(0.92000000000000004D * (double)dimension.width), (int)(0.69999999999999996D * (double)dimension.height));
            int ai[] = new int[3];
            int ai1[] = new int[3];
            int j1 = fontmetrics.stringWidth("TIME ");
            g.drawLine((int)(0.92000000000000004D * (double)dimension.width), (int)(0.69999999999999996D * (double)dimension.height) + i, (int)(0.92000000000000004D * (double)dimension.width) + j1 / 2, (int)(0.69999999999999996D * (double)dimension.height) + i);
            ai[0] = (int)(0.92000000000000004D * (double)dimension.width) + j1 / 2;
            ai[1] = (int)(0.92000000000000004D * (double)dimension.width) + j1 / 2 + 6;
            ai[2] = (int)(0.92000000000000004D * (double)dimension.width) + j1 / 2;
            ai1[0] = ((int)(0.69999999999999996D * (double)dimension.height) + i) - 3 * j;
            ai1[1] = (int)(0.69999999999999996D * (double)dimension.height) + i;
            ai1[2] = (int)(0.69999999999999996D * (double)dimension.height) + i + 3 * j;
            g.fillPolygon(ai, ai1, 3);
            g.drawString("Packet Timer before time out.", (int)(0.65000000000000002D * (double)dimension.width), byte0);
            g.drawString("Packet Timer after time out.", (int)(0.65000000000000002D * (double)dimension.width), byte0 + i);
            g.drawString("Packet", (int)(0.65000000000000002D * (double)dimension.width), byte0 + 2 * i);
            g.drawString("Acknowledgement", (int)(0.65000000000000002D * (double)dimension.width), byte0 + 3 * i);
            g.drawString("Lost/corrupted Packet or ACK.", (int)(0.65000000000000002D * (double)dimension.width), byte0 + 4 * i);
            g.fillRect((int)(0.65000000000000002D * (double)dimension.width) - 14, byte0 - 2 * j, 7, 2);
            g.setFont(new Font("TimesRoman", 0, 10));
            g.drawString("0", i1 + 4 * k, byte0 + i + 2 * j);
            g.drawString("Time Out", i1 + 9 * k, byte0 + i + 2 * j);
            g.setColor(Color.red);
            g.fillRect((int)(0.65000000000000002D * (double)dimension.width) - 14, (byte0 + i) - 2 * j, 7, 2);
            g.fillRect((int)(0.65000000000000002D * (double)dimension.width) - 14, (byte0 + 4 * i) - l / 2, 7, 2 * j);
            g.setColor(Color.blue);
            g.fillRect((int)(0.65000000000000002D * (double)dimension.width) - 14, (byte0 + 2 * i) - l / 2, 7, 2 * j);
            g.setColor(Color.green);
            g.fillRect((int)(0.65000000000000002D * (double)dimension.width) - 14, (byte0 + 3 * i) - l / 2, 7, 2 * j);
            g.setColor(getBackground());
            g.fill3DRect((int)(0.050000000000000003D * (double)dimension.width), (int)(0.45000000000000001D * (double)dimension.height) + j, (int)(0.87D * (double)dimension.width), 8, true);
            g.fill3DRect((int)(0.050000000000000003D * (double)dimension.width), (int)(0.97999999999999998D * (double)dimension.height) - (i + 2 * j), (int)(0.87D * (double)dimension.width), 8, true);
            int k1 = i1 + 4 * k;
            int l1 = byte0 + i + 3 * j;
            int i2 = byte0 + 5 * i + 2 * j;
            int j2 = i1 + 10 * k;
            Rectangle rectangle = new Rectangle(k1, l1, j2 - k1, i2 - l1);
            k1 = (int)(0.050000000000000003D * (double)dimension.width);
            l1 = (int)(0.45000000000000001D * (double)dimension.height) + j + 8;
            i2 = (int)(0.97999999999999998D * (double)dimension.height) - (i + 2 * j);
            j2 = (int)(0.92000000000000004D * (double)dimension.width);
            Rectangle rectangle1 = new Rectangle(k1, l1, j2 - k1, i2 - l1);
            if(currAnimation.getType() == 0)
            {
                g.setColor(getBackground());
                g.fillRect(rectangle.x, rectangle.y, rectangle.width, rectangle.height);
                g.fillRect(rectangle1.x, rectangle1.y, rectangle1.width, rectangle1.height);
                return;
            }
            if(appletContainer.sim.isDone())
            {
                appletContainer.resetSim();
                return;
            }
            Image image = createImage(rectangle1.width, rectangle1.height);
            Graphics g1 = image.getGraphics();
            g1.setColor(getBackground());
            g1.fillRect(0, 0, rectangle1.width, rectangle1.height);
            g1.setFont(new Font("TimesRoman", 0, 10));
            appletContainer.sim.update(g, g1, rectangle1, rectangle, i);
            g.drawImage(image, rectangle1.x, rectangle1.y, this);
        }
    }

    private boolean isCanvasSizeOK(Graphics g, Dimension dimension)
    {
        if(dimension.width < 600 || dimension.height < 226)
        {
            g.setColor(getBackground());
            g.fillRect(0, 0, dimension.width, dimension.height);
            g.setColor(Color.black);
            g.drawString("The Applet Window is too small to run simulation.", (int)(0.29999999999999999D * (double)dimension.width), (int)(0.5D * (double)dimension.height));
            return false;
        } else
        {
            return true;
        }
    }

    public Dimension minimumSize()
    {
        return new Dimension(600, 226);
    }

    public Dimension preferredSize()
    {
        return new Dimension(600, 250);
    }

    private AnimationType currAnimation;
    private ARQSimulation appletContainer;
}



class FindFrame
{

    public static Frame getFrame(Component component)
    {
        Object obj = component;
        Frame frame = null;
        for(; obj != null; obj = ((Component) (obj)).getParent())
        {
            if(!(obj instanceof Frame))
                continue;
            frame = (Frame)obj;
            break;
        }

        return frame;
    }

    FindFrame()
    {
    }
}



class FramedPanel extends Panel
{

    public FramedPanel(DrawingArea drawingarea)
    {
        setLayout(new GridLayout(1, 1));
        add(drawingarea);
        validate();
    }

    public Insets insets()
    {
        return new Insets(4, 4, 4, 4);
    }

    public void paint(Graphics g)
    {
        Dimension dimension = size();
        java.awt.Color color = getBackground();
        g.setColor(color);
        g.draw3DRect(0, 0, dimension.width - 1, dimension.height - 1, true);
        g.draw3DRect(3, 3, dimension.width - 7, dimension.height - 7, false);
    }
}


class MsgFrame
{

    public MsgFrame(int i, int j, int k, int l)
    {
        atReceiver = false;
        atSender = false;
        leftSender = false;
        leftReceiver = false;
        done = false;
        finishing = false;
        msgNum = String.valueOf(l);
        numOfMsg = l;
        msgType = i;
        framewidth = j;
        frameheight = k;
    }

    public void update(Graphics g, int i, int j)
    {
        int ai[] = new int[4];
        int ai1[] = new int[4];
        switch(msgType)
        {
        case 0: // '\0'
        case 1: // '\001'
            if(!leftReceiver)
            {
                initialHeight = initialHeight + 2;
                if(initialHeight >= frameheight)
                    leftReceiver = true;
                ai[0] = i;
                ai[1] = ai[0] + initialHeight / 2;
                ai[2] = ai[1] + framewidth;
                ai[3] = ai[0] + framewidth;
                ai1[0] = ai1[3] = j;
                ai1[1] = ai1[2] = ai1[0] - initialHeight;
                g.setColor(Color.green);
                g.fillPolygon(ai, ai1, 4);
                return;
            }
            currxpos++;
            currypos = currypos + 2;
            ai[0] = i + currxpos;
            ai[1] = ai[0] + frameheight / 2;
            ai[2] = ai[1] + framewidth;
            ai[3] = ai[0] + framewidth;
            ai1[0] = ai1[3] = j - currypos;
            ai1[1] = ai1[2] = ai1[0] - frameheight;
            int k = ai[0] + framewidth / 2;
            int i1 = ai1[0] + 10;
            if(!finishing)
                finalxpos = currxpos;
            if(msgType == 1)
                if(ai1[1] <= j - (int)(0.69999999999999996D * (double)j))
                {
                    ai1[1] = ai1[2] = j - (int)(0.69999999999999996D * (double)j);
                    finishing = true;
                    ai[1] = i + finalxpos + frameheight / 2;
                    ai[2] = ai[1] + framewidth;
                    if(ai1[0] <= j - (int)(0.69999999999999996D * (double)j))
                    {
                        done = true;
                        return;
                    } else
                    {
                        g.setColor(Color.black);
                        g.drawString(msgNum, k, i1);
                        g.setColor(Color.red);
                        g.fillPolygon(ai, ai1, 4);
                        return;
                    }
                } else
                {
                    g.setColor(Color.black);
                    g.drawString(msgNum, k, i1);
                    g.setColor(Color.green);
                    g.fillPolygon(ai, ai1, 4);
                    return;
                }
            if(ai1[1] <= 0)
            {
                ai1[1] = ai1[2] = 0;
                finishing = true;
                ai[1] = i + finalxpos + frameheight / 2;
                ai[2] = ai[1] + framewidth;
                if(ai1[0] <= 0)
                {
                    atSender = true;
                    done = true;
                    return;
                } else
                {
                    g.setColor(Color.black);
                    g.drawString(msgNum, k, i1);
                    g.setColor(Color.green);
                    g.fillPolygon(ai, ai1, 4);
                    return;
                }
            } else
            {
                g.setColor(Color.black);
                g.drawString(msgNum, k, i1);
                g.setColor(Color.green);
                g.fillPolygon(ai, ai1, 4);
                return;
            }

        case 2: // '\002'
        case 3: // '\003'
            if(!leftSender)
            {
                initialHeight = initialHeight + 2;
                if(initialHeight >= frameheight)
                    leftSender = true;
                ai[0] = i;
                ai[1] = ai[0] + initialHeight / 2;
                ai[2] = ai[1] + framewidth;
                ai[3] = ai[0] + framewidth;
                ai1[0] = ai1[3] = 0;
                ai1[1] = ai1[2] = ai1[0] + initialHeight;
                g.setColor(Color.blue);
                g.fillPolygon(ai, ai1, 4);
                return;
            }
            currxpos++;
            currypos = currypos + 2;
            ai[0] = i + currxpos;
            ai[1] = ai[0] + frameheight / 2;
            ai[2] = ai[1] + framewidth;
            ai[3] = ai[0] + framewidth;
            ai1[0] = ai1[3] = currypos;
            ai1[1] = ai1[2] = ai1[0] + frameheight;
            int l = ai[0] + framewidth / 2;
            int j1 = ai1[0] - 2;
            if(!finishing)
                finalxpos = currxpos;
            if(msgType == 3)
                if(ai1[1] >= (int)(0.69999999999999996D * (double)j))
                {
                    ai1[1] = ai1[2] = (int)(0.69999999999999996D * (double)j);
                    finishing = true;
                    ai[1] = i + finalxpos + frameheight / 2;
                    ai[2] = ai[1] + framewidth;
                    if(ai1[0] >= (int)(0.69999999999999996D * (double)j))
                    {
                        done = true;
                        return;
                    } else
                    {
                        g.setColor(Color.black);
                        g.drawString(msgNum, l, j1);
                        g.setColor(Color.red);
                        g.fillPolygon(ai, ai1, 4);
                        return;
                    }
                } else
                {
                    g.setColor(Color.black);
                    g.drawString(msgNum, l, j1);
                    g.setColor(Color.blue);
                    g.fillPolygon(ai, ai1, 4);
                    return;
                }
            if(ai1[1] >= j)
            {
                ai1[1] = ai1[2] = j;
                finishing = true;
                ai[1] = i + finalxpos + frameheight / 2;
                ai[2] = ai[1] + framewidth;
                if(ai1[0] >= j)
                {
                    done = true;
                    atReceiver = true;
                    return;
                } else
                {
                    g.setColor(Color.black);
                    g.drawString(msgNum, l, j1);
                    g.setColor(Color.blue);
                    g.fillPolygon(ai, ai1, 4);
                    return;
                }
            } else
            {
                g.setColor(Color.black);
                g.drawString(msgNum, l, j1);
                g.setColor(Color.blue);
                g.fillPolygon(ai, ai1, 4);
                return;
            }
        }
    }

    public boolean isDone()
    {
        return done;
    }

    public int getMsgNum()
    {
        return numOfMsg;
    }

    public boolean isAtReceiver()
    {
        return atReceiver;
    }

    public boolean isAtSender()
    {
        return atSender;
    }

    public boolean hasLeftReceiver()
    {
        return leftReceiver;
    }

    public boolean hasLeftSender()
    {
        return leftSender;
    }

    public static final int ACK = 0;
    public static final int LOSTACK = 1;
    public static final int PKT = 2;
    public static final int LOSTPKT = 3;
    private boolean atReceiver;
    private boolean atSender;
    private boolean leftSender;
    private boolean leftReceiver;
    private boolean done;
    private String msgNum;
    private int numOfMsg;
    private int msgType;
    private int initialHeight;
    private boolean finishing;
    private int currxpos;
    private int currypos;
    private int finalxpos;
    private int framewidth;
    private int frameheight;
}



class Queue
{

    public Queue(int i)
    {
        queue = new int[i + 1];
        size = i + 1;
    }

    public boolean isEmpty()
    {
        return head == tail;
    }

    public void insert(int i)
    {
        tail = (tail + 1) % size;
        queue[tail] = i;
    }

    public int remove()
    {
        head = (head + 1) % size;
        return queue[head];
    }

    public int viewHead()
    {
        return queue[(head + 1) % size];
    }

    private int queue[];
    private int head;
    private int tail;
    private int size;
}


class Receiver
{

    public Receiver(SimARQ simarq, AnimationType animationtype, int i, int j)
    {
        transmitting = false;
        alreadyTransmittedAck1 = false;
        nextPktExpected = 1;
        nextAckId = 1;
        animationType = animationtype;
        framewidth = i;
        frameheight = j;
        simulation = simarq;
        baseXpos = new int[5];
        pktReceived = new boolean[5];
        ack = new MsgFrame[5];
        for(int k = 0; k < pktReceived.length; k++)
            pktReceived[k] = false;

        receivedPkts = new Queue(4);
    }

    public void deallocate()
    {
        animationType = null;
        simulation = null;
        for(int i = 0; i < ack.length; i++)
            ack[i] = null;

        ack = null;
        receivedPkts = null;
        pktReceived = null;
        baseXpos = null;
    }

    public void update(Graphics g, int i, Sender sender)
    {
        switch(animationType.getType())
        {
        case 1: // '\001'
        case 7: // '\007'
        case 8: // '\b'
            for(int j = 1; j < pktReceived.length; j++)
                if(pktReceived[j])
                {
                    if(ack[j] == null)
                        if(animationType.getType() == 7 && j == 1)
                        {
                            if(alreadyTransmittedAck1)
                            {
                                ack[j] = new MsgFrame(0, framewidth, frameheight, j);
                            } else
                            {
                                ack[j] = new MsgFrame(1, framewidth, frameheight, j);
                                alreadyTransmittedAck1 = true;
                            }
                        } else
                        {
                            ack[j] = new MsgFrame(0, framewidth, frameheight, j);
                        }
                    if(ack[j].isAtSender())
                    {
                        sender.deliverAck(j);
                        ack[j] = null;
                        pktReceived[j] = false;
                    } else
                    if(ack[j].isDone())
                    {
                        ack[j] = null;
                        pktReceived[j] = false;
                    } else
                    if(ack[j].hasLeftReceiver())
                        ack[j].update(g, baseXpos[j], i);
                    else
                    if(transmitting)
                    {
                        if(ackNumTransmitting == j && simulation.getTimerTics() > baseXpos[j])
                        {
                            ack[j].update(g, baseXpos[j], i);
                            if(ack[j].hasLeftReceiver())
                            {
                                transmitting = false;
                                ackNumTransmitting = 0;
                            }
                        }
                    } else
                    {
                        transmitting = true;
                        ackNumTransmitting = j;
                        baseXpos[j] = simulation.getTimerTics() + 3;
                    }
                }

            return;

        case 2: // '\002'
        case 3: // '\003'
        case 4: // '\004'
            if(!receivedPkts.isEmpty())
            {
                int k = receivedPkts.remove();
                pktReceived[k] = true;
                if(k == nextPktExpected)
                {
                    for(int i1 = nextPktExpected; i1 < pktReceived.length; i1++)
                    {
                        if(!pktReceived[i1])
                            break;
                        nextPktExpected = i1 + 1;
                    }

                }
                if(animationType.getType() == 4 && k == 1)
                {
                    if(alreadyTransmittedAck1)
                    {
                        ack[nextAckId] = new MsgFrame(0, framewidth, frameheight, nextPktExpected);
                        nextAckId++;
                    } else
                    {
                        ack[nextAckId] = new MsgFrame(1, framewidth, frameheight, nextPktExpected);
                        nextAckId++;
                        alreadyTransmittedAck1 = true;
                    }
                } else
                {
                    ack[nextAckId] = new MsgFrame(0, framewidth, frameheight, nextPktExpected);
                    nextAckId++;
                }
            }
            for(int l = 1; l < ack.length; l++)
                if(ack[l] != null)
                    if(ack[l].isAtSender())
                    {
                        sender.deliverAck(ack[l].getMsgNum());
                        ack[l] = null;
                    } else
                    if(ack[l].isDone())
                        ack[l] = null;
                    else
                    if(ack[l].hasLeftReceiver())
                        ack[l].update(g, baseXpos[l], i);
                    else
                    if(transmitting)
                    {
                        if(ackNumTransmitting == l && simulation.getTimerTics() > baseXpos[l])
                        {
                            ack[l].update(g, baseXpos[l], i);
                            if(ack[l].hasLeftReceiver())
                            {
                                transmitting = false;
                                ackNumTransmitting = 0;
                            }
                        }
                    } else
                    {
                        transmitting = true;
                        ackNumTransmitting = l;
                        baseXpos[l] = simulation.getTimerTics() + framewidth;
                    }

            return;

        case 5: // '\005'
        case 6: // '\006'
        default:
            return;
        }
    }

    public void deliverPkt(int i)
    {
        if(animationType.getType() == 4 || animationType.getType() == 3 || animationType.getType() == 2)
        {
            receivedPkts.insert(i);
            return;
        } else
        {
            pktReceived[i] = true;
            return;
        }
    }

    private AnimationType animationType;
    private int framewidth;
    private int frameheight;
    private SimARQ simulation;
    private boolean transmitting;
    private int ackNumTransmitting;
    private boolean alreadyTransmittedAck1;
    private int baseXpos[];
    private MsgFrame ack[];
    private boolean pktReceived[];
    private int nextPktExpected;
    private Queue receivedPkts;
    private int nextAckId;
}



class Sender
{

    public Sender(SimARQ simarq, Color color, AnimationType animationtype, int i, int j, float f, int k, 
            int l)
    {
        done = false;
        transmitting = false;
        nextFrameToTransmit = 1;
        background = color;
        animationType = animationtype;
        bandwidth = i;
        framesize = j;
        framedelay = f;
        framewidth = k;
        frameheight = l;
        simulation = simarq;
        pkt = new MsgFrame[5];
        timer = new Timer[5];
        baseXpos = new int[5];
        acksReceived = new Queue(4);
        retransmissionList = new Queue(4);
        if(animationtype.getType() == 1 || animationtype.getType() == 7 || animationtype.getType() == 8)
        {
            MaxWindowSize = 1;
            TotalNumPkts = 2;
            return;
        } else
        {
            MaxWindowSize = 3;
            TotalNumPkts = 4;
            return;
        }
    }

    public void deallocate()
    {
        animationType = null;
        simulation = null;
        for(int i = 0; i < pkt.length; i++)
        {
            pkt[i] = null;
            timer[i] = null;
        }

        pkt = null;
        acksReceived = null;
        retransmissionList = null;
        background = null;
        baseXpos = null;
    }

    public void update(Graphics g, Graphics g1, int i, Rectangle rectangle, int j, Receiver receiver)
    {
        switch(animationType.getType())
        {
        case 1: // '\001'
        case 7: // '\007'
        case 8: // '\b'
            if(!acksReceived.isEmpty())
            {
                int k = acksReceived.remove();
                int j1;
                for(j1 = retransmissionList.remove(); j1 != k; j1 = retransmissionList.remove())
                {
                    pkt[j1] = new MsgFrame(2, framewidth, frameheight, j1);
                    timer[j1].reset(g);
                    retransmissionList.insert(j1);
                }

                numOutstandingPkts--;
                timer[j1].reset(g);
                timer[j1] = null;
                if(k == TotalNumPkts)
                    done = true;
            }
            break;

        case 2: // '\002'
        case 3: // '\003'
        case 4: // '\004'
            if(!acksReceived.isEmpty())
            {
                int l = acksReceived.remove();
                int k1 = retransmissionList.viewHead();
                if(l == k1)
                {
                    pkt[k1] = new MsgFrame(2, framewidth, frameheight, k1);
                    timer[k1].reset(g);
                    retransmissionList.remove();
                    retransmissionList.insert(k1);
                } else
                {
                    for(; l > k1 && !retransmissionList.isEmpty(); k1 = retransmissionList.viewHead())
                    {
                        retransmissionList.remove();
                        timer[k1].reset(g);
                        timer[k1] = null;
                        numOutstandingPkts--;
                    }

                }
                if(l == TotalNumPkts + 1)
                    done = true;
            }
            break;
        }
        if(numOutstandingPkts < MaxWindowSize && nextFrameToTransmit <= TotalNumPkts)
        {
            if(animationType.getType() == 3 && nextFrameToTransmit == 2 || animationType.getType() == 8 && nextFrameToTransmit == 1)
                pkt[nextFrameToTransmit] = new MsgFrame(3, framewidth, frameheight, nextFrameToTransmit);
            else
                pkt[nextFrameToTransmit] = new MsgFrame(2, framewidth, frameheight, nextFrameToTransmit);
            timer[nextFrameToTransmit] = new Timer(animationType, nextFrameToTransmit, background, j, rectangle);
            retransmissionList.insert(nextFrameToTransmit);
            numOutstandingPkts++;
            nextFrameToTransmit++;
        }
        for(int i1 = 1; i1 < pkt.length; i1++)
            if(pkt[i1] != null)
            {
                if(pkt[i1].isAtReceiver())
                {
                    pkt[i1] = null;
                    receiver.deliverPkt(i1);
                    timer[i1].update(g);
                } else
                if(pkt[i1].isDone())
                {
                    pkt[i1] = null;
                    timer[i1].update(g);
                } else
                if(pkt[i1].hasLeftSender())
                {
                    pkt[i1].update(g1, baseXpos[i1], i);
                    timer[i1].update(g);
                } else
                if(transmitting)
                {
                    if(pktNumTransmitting == i1 && simulation.getTimerTics() > baseXpos[i1])
                    {
                        pkt[i1].update(g1, baseXpos[i1], i);
                        timer[i1].update(g);
                        if(pkt[i1].hasLeftSender())
                        {
                            transmitting = false;
                            pktNumTransmitting = 0;
                        }
                    }
                } else
                {
                    transmitting = true;
                    pktNumTransmitting = i1;
                    if(animationType.getType() == 7 || animationType.getType() == 8)
                        baseXpos[i1] = simulation.getTimerTics() + 3;
                    else
                        baseXpos[i1] = simulation.getTimerTics() + framewidth;
                }
            } else
            if(timer[i1] != null)
                if(timer[i1].isTimerExpired())
                {
                    timer[i1].reset(g);
                    pkt[i1] = new MsgFrame(2, framewidth, frameheight, i1);
                } else
                {
                    timer[i1].update(g);
                }

    }

    public void deliverAck(int i)
    {
        acksReceived.insert(i);
    }

    public boolean isDone()
    {
        return done;
    }

    private boolean done;
    private Color background;
    private AnimationType animationType;
    private int bandwidth;
    private int framesize;
    private float framedelay;
    private int framewidth;
    private int frameheight;
    private int baseXpos[];
    private SimARQ simulation;
    private MsgFrame pkt[];
    private Timer timer[];
    private Queue acksReceived;
    private Queue retransmissionList;
    private boolean transmitting;
    private int pktNumTransmitting;
    private int numOutstandingPkts;
    private int MaxWindowSize;
    private int nextFrameToTransmit;
    private int TotalNumPkts;
}



class SimARQ
{

    public SimARQ(SynCounter syncounter, Color color, AnimationType animationtype, int i, int j, float f)
    {
        done = false;
        MAXFRAMEHEIGHT = 35;
        MINFRAMEHEIGHT = 3;
        FRAMEHEIGHTRANGE = 32F;
        MAXFRAMEWIDTH = 35;
        MINFRAMEWIDTH = 5;
        FRAMEWIDTHRANGE = 30F;
        MAXFRAMESIZE = 4500F;
        MINBANDWIDTH = 10F;
        MAXBANDWIDTH = 100F;
        MAXTRANSMITTIME = MAXFRAMESIZE / MINBANDWIDTH;
        frameNum = syncounter;
        int k = MINFRAMEWIDTH + (int)((FRAMEWIDTHRANGE * (float)i) / MAXBANDWIDTH);
        float f1 = (float)j / (float)i;
        int l = MINFRAMEHEIGHT + (int)((FRAMEHEIGHTRANGE * f1) / MAXTRANSMITTIME);
        sender = new Sender(this, color, animationtype, i, j, f, k, l);
        int i1 = MINFRAMEWIDTH + (int)((FRAMEWIDTHRANGE * (float)i) / MAXBANDWIDTH);
        float f2 = 64F;
        float f3 = f2 / (float)i;
        int j1 = MINFRAMEHEIGHT + (int)((FRAMEHEIGHTRANGE * f3) / MAXTRANSMITTIME);
        receiver = new Receiver(this, animationtype, i1, j1);
    }

    public void deallocate()
    {
        sender.deallocate();
        receiver.deallocate();
        sender = null;
        receiver = null;
        frameNum = null;
    }

    public void update(Graphics g, Graphics g1, Rectangle rectangle, Rectangle rectangle1, int i)
    {
        done = sender.isDone();
        simulationTime = frameNum.getCounter();
        if(!done)
        {
            sender.update(g, g1, rectangle.height, rectangle1, i, receiver);
            receiver.update(g1, rectangle.height, sender);
        }
    }

    public int getTimerTics()
    {
        return simulationTime;
    }

    public boolean isDone()
    {
        return done;
    }

    private boolean done;
    private Sender sender;
    private Receiver receiver;
    private int simulationTime;
    private SynCounter frameNum;
    private int MAXFRAMEHEIGHT;
    private int MINFRAMEHEIGHT;
    private float FRAMEHEIGHTRANGE;
    private int MAXFRAMEWIDTH;
    private int MINFRAMEWIDTH;
    private float FRAMEWIDTHRANGE;
    private float MAXFRAMESIZE;
    private float MINBANDWIDTH;
    private float MAXBANDWIDTH;
    private float MAXTRANSMITTIME;
}



class SynCounter
{

    public synchronized void resetCounter()
    {
        readCounter = false;
        counter = 0;
    }

    public synchronized int getCounter()
    {
        readCounter = true;
        notify();
        return counter;
    }

    public synchronized void incrementCounter()
    {
        while(!readCounter) 
            try
            {
                wait();
            }
            catch(InterruptedException _ex) { }
        counter++;
        readCounter = false;
    }

    SynCounter()
    {
        readCounter = false;
    }

    private int counter;
    private boolean readCounter;
}



class Timer
{

    public Timer(AnimationType animationtype, int i, Color color, int j, Rectangle rectangle)
    {
        expired = false;
        timerNum = i;
        animationType = animationtype;
        background = color;
        area = rectangle;
        switch(i)
        {
        case 1: // '\001'
            ycoor = rectangle.y + 2;
            return;

        case 2: // '\002'
        case 3: // '\003'
        case 4: // '\004'
            ycoor = rectangle.y + (i - 1) * j + 2;
            return;
        }
    }

    public boolean isTimerExpired()
    {
        return expired;
    }

    public void reset(Graphics g)
    {
        timerTics = 0.0F;
        expired = false;
        g.setColor(background);
        g.fillRect(area.x, ycoor, area.width, 3);
    }

    public void update(Graphics g)
    {
        if(animationType.getType() == 7 || animationType.getType() == 8)
            timerTics = timerTics + 0.5F;
        else
            timerTics = timerTics + 0.2F;
        if(timerTics >= (float)area.width)
            expired = true;
        if(expired)
        {
            g.setColor(Color.red);
            g.fillRect(area.x, ycoor, area.width, 3);
            return;
        } else
        {
            g.setColor(Color.black);
            g.fillRect(area.x, ycoor, (int)timerTics, 3);
            return;
        }
    }

    private Color background;
    private AnimationType animationType;
    private float timerTics;
    private boolean expired;
    private int timerNum;
    private int ycoor;
    private Rectangle area;
}



