/* 

			       Simulation of Routing Algorithms.

 */


import java.applet.*;
import java.awt.*;
import java.awt.event.*;

public class fixed extends Applet implements ActionListener, Runnable
 {
  int n, source, dest;
  static int algo = 2;                   // algorithm selected (0 - Fixed; 1 - Flooding).
  boolean done;                          // source, destination values entered? 
  int rtable[][] = new int[8][8];        // routing table.
  int x[] = {40, 140, 140, 240, 340, 440, 440, 540};    // x-position of node.
  int y[] = {230, 130, 330, 230, 230, 130, 330, 230};   // y-position of node.
  int dist[][] = {{0, 2, 6, 0, 0, 0, 0, 0},             // distances between nodes.
  			{2, 0, 0, 2, 0, 7, 0, 0},             // '0' indicates no link. 
  			{6, 0, 0, 1, 0, 0, 4, 0},   
  			{0, 2, 1, 0, 2, 0, 0, 0},
  			{0, 0, 0, 2, 0, 3, 2, 0},
  			{0, 7, 0, 0, 3, 0, 0, 3},
  			{0, 0, 4, 0, 2, 0, 0, 2},
  			{0, 0, 0, 0, 0, 3, 2, 0}
  		     };
  int prev[] = new int[8];             // stores node before current one in path. 
  int length[] = new int[8];           // stores length of path till current node. 
  int path[] = new int[8];             // stores path from source to destination.
  boolean label[] = new boolean[8];    // indicates whether node is in path or not.
  Thread wait;                         // thread for simulation. 
  static Button table, option, start, pause, simhelp;
  Canvas1 canvas;
  Window rtabl;
  Win nodes;
  Options options;
	Help help = new Help("Help");
 
  public void init()
   {
    int i, j;   
    setBackground(Color.lightGray);
    setSize(640, 460);
    add(canvas = new Canvas1());
    canvas.setSize(640, 380);
    rtabl = new Window("Routing Table");
    nodes = new Win("Source and Destination Nodes");
    options = new Options("Parameters");
    add(table = new Button(" Routing Table"));
    add(option = new Button("   Options   "));
    add(start = new Button("    Start     "));
    add(pause = new Button("    Pause     "));
    add(simhelp = new Button("     Help      "));
    table.addActionListener(this);
    option.addActionListener(this);
    start.addActionListener(this);
    pause.addActionListener(this);
    simhelp.addActionListener(this);
    for (i = 0; i < 8; i++)
     for (j = 0; j < 8; j++)
      rtable[i][j] = 8;        // initialising to NULL (nodes are from 0 to 7).
    table.setEnabled(false);
//    start.setEnabled(false);
    pause.setEnabled(false);
    prepareTable();                   // prepares routing table.
    rtabl.canvas.copyRtable(rtable);  // passes routing table to canvas class for display.
   }
  
  public void actionPerformed(ActionEvent event)
   {
    if (event.getSource() == table) showTable();
    if (event.getSource() == option) setParam();
    try
     {
      if (event.getSource() == start) startOrStop();
     }
     catch(NullPointerException e)
      {
       startOrStop();
      }
    try  
     {
      if (event.getSource() == pause) pauseOrResume();
     }
     catch(NullPointerException e)
      {
       pauseOrResume();
      }
		if (event.getSource() == simhelp) 
		  {
			 help.setSize(590, 410);
			 help.show();	
			}
   }
   
  private void showTable()       // displays the routing table.
   {
    rtabl.setSize(340, 340);
    rtabl.show();
   }
   
  private void setParam()        // brings up the options window.
   {
    options.delay1 = options.delay;   // store initial values incase 
    options.algo1 = algo;             // changes are not saved.
    options.setSize(300, 150);
    options.show();    
   }
   
  private void startOrStop()
   {
    if (start.getLabel() == "Start")   // Start button (simulation stopped).
     {
      nodes.ok.setEnabled(false);     
      nodes.source = nodes.dest = 0;   // initialise values of source and destination.
      nodes.done = false;              // initially values not entered.
      nodes.setSize(300, 200);
      nodes.show();
      start();
     } 
     else              // Stop button (simulation in progress).
      {
       if (algo == 0) canvas.packet.xpos = canvas.packet.ypos = 0;
       stop();
       if (algo == 1)
        {
         for (int i = 0; i <= canvas.packets; i++)
          if (canvas.threads[i].isAlive()) canvas.threads[i].stop();
         for (int i = 0; i < 8; i++)
          for (int j = 0; j < 8; j++)
           Packet.sent[i][j] = false;
        }  
      } 
   }

  public void pauseOrResume()
   { 
    int i;
    if (pause.getLabel() == "Pause")   // Pause button (simulation in progress).
     {
      if (algo == 0) wait.suspend();
       else if (algo == 1)
        {
         for (i = 0; i <= canvas.packets; i++)
          if (canvas.threads[i].isAlive()) 
           try
            {
             canvas.threads[i].suspend();
            }
            catch(NullPointerException e)
             {
              canvas.threads[i].suspend();
             }
        }     
      pause.setLabel("Resume");
     }
     else      // Resume button (simulation has been paused).
      {
       if (algo == 0) wait.resume();
        else if (algo == 1)
         {
          for (i = 0; i <= canvas.packets; i++)
          if (canvas.threads[i].isAlive()) 
           try
            {
             canvas.threads[i].resume();
            }
            catch(NullPointerException e)
             {
              canvas.threads[i].resume();
             }
         }   
       pause.setLabel("Pause");
      }
   }
   
  private void prepareTable()          // fills up the routing table.   
   {
    int i, j, k, m, tmp1, tmp2, nodes;   
    for (i = 0; i < 8; i++)                  // for each node 'i' in network,
     {                                       
      for (j = 0; j < 8; j++)                // find shortest path to every other node 'j'.  
       {
        if ((i != j) && (rtable[i][j] == 8)) // don't check for path from node to itself
         {						   // rtable = 8 checks if shortest path has already been found using optimality principle.	
          nodes = shortestPath(i, j);        // nodes is no. of nodes in the shortest path.
          for (m = 0; m < (nodes - 1); m++)  // use shortest path found to fill routing table, also using optimality principle.
           {
            tmp1 = path[m];
            for (k = (m + 1); k < nodes; k++)
             {
              tmp2 = path[k];
              if (rtable[tmp1][tmp2] == 8)
               {
                rtable[tmp1][tmp2] = path[m + 1];
                rtable[tmp2][tmp1] = path[k - 1];
               }
             }
           }  
         }
       }
     }  
   }
   
  private int shortestPath(int s, int t)  // uses Dijkstra's Shortest Path Algorithm.
   {
    int k, m, min;
    for (k = 0; k < 8; k++)           // initialise values.
     {
      prev[k] = -1;                   // node just before node 'k' in path.
      length[k] = Integer.MAX_VALUE;  // length of path to node 'k'.
      label[k] = false;               // true means that node 'k' has been finalised as part of the shortest path.
     }
    length[t] = 0;          // Algorithm implementation begins here.
    label[t] = true;
    k = t;
    do
     {
      for (m = 0; m < 8; m++)
       {
        if ((dist[k][m] != 0) && (label[m] == false))
         {
          if ((length[k] + dist[k][m]) < length[m])
           {
            prev[m] = k;
            length[m] = length[k] + dist[k][m];
           }
         }  
       }
      k = 0; 
      min = Integer.MAX_VALUE;
      for (m = 0; m < 8; m++)
       {
        if ((label[m] == false) && (length[m] < min))
         { 
          min = length[m];
          k = m;
         }
       }
      label[k] = true; 
     }  
    while (k != s);
    m = 0;
    k = s;
    do
     {
      path[m++] = k;
      k = prev[k];
     }
    while (k >= 0);
    return(m);       // 'm' is no. of nodes in shortest path.
   }
  
  public void start()         // thread which handles simulation.
   {
    wait = new Thread(this);  
    wait.start();
   }
   
  public void run()
   {
    while ((done = nodes.getDone()) == false);   // wait for user to enter source and destination nodes.
    source = nodes.getSource();                // get values of source
    dest = nodes.getDest();                    // and destination. 
    Graphics g = canvas.getGraphics();
    canvas.activateNodes(g, source, dest);     // mark source and destination nodes.
    start.setLabel("Stop");   
    pause.setEnabled(true);
    option.setEnabled(false);
    canvas.sendInfo(source, dest, rtable, options.delay);  // simulate the transfer of packet.
    stop();
   } 
   
  public void stop()     // stop simulation.
   {
    canvas.redorblue[source] = canvas.redorblue[dest] = false;
    canvas.repaint();
    start.setLabel("Start");
    option.setEnabled(true);
    pause.setLabel("Pause");
    pause.setEnabled(false);
    wait.stop();
   }
   
  public int getx(int j)
   {
    return(x[j]);
   }
   
  public int gety(int j)
   {
    return(y[j]);
   }
   
  public int getDist(int i, int j)
   {
    return(dist[i][j]);
   }

  public String getAppletInfo()
   {
    return "Project Title :- Tools for Network and Protocol Simulation"
	     + "\r\n\n";
   }
 }  
 
class Canvas1 extends Canvas implements Runnable
 {
  boolean redorblue[] = new boolean[8];   // indicates whether node is in Red or Blue (source and destination are in Red). 
  int packets;    // keeps track of number of packets/threads created in flooding.
  fixed f; 
  Packet packet;
  Packet fpacket[] = new Packet[200];
  Thread threads[] = new Thread[200];
  Font font;
  FontMetrics fontm;
  String title = "Simulation of Routing Algorithms";

// colors makes each node send out packets in different colors for easy understanding of packet flow in flooding.  

  Color colors[] = { Color.darkGray, Color.green, Color.pink, Color.magenta,
                     Color.black, Color.white, Color.yellow, Color.cyan }; 
  
  Canvas1()
   {
    super();
    f = new fixed();
    packet = new Packet();
    for (int i = 0; i < 8; i++)   // set all nodes to Blue.
     redorblue[i] = false;
   }
   
  public void paint(Graphics g)
   {
    int i, j, x, y, xtmp, ytmp;
    g.setFont(font = new Font("Times New Roman", Font.PLAIN |  Font.BOLD, 16));
    fontm = getFontMetrics(font);
    Dimension d = getSize();
    g.setColor(Color.black);
    if (fixed.algo == 0) title = "Simulation of Fixed Routing Algorithm";
    if (fixed.algo == 1) title = "Simulation of Flooding Algorithm";
    g.drawString(title, (d.width - fontm.stringWidth(title)) / 2, 15);
    g.setFont(font = new Font("Times New Roman", Font.PLAIN, 10));
    if (fixed.algo == 0)    // draws the key for fixed routing.
     {
      g.setColor(Color.lightGray);
      g.fill3DRect(5, 40, 150, 50, true);
      g.setColor(Color.pink);
      g.fill3DRect(25, 53, 6, 7, true);
      g.setColor(Color.darkGray);
      g.fill3DRect(15, 72, 16, 7, true);
      g.setColor(Color.black);
      g.drawString("Destination address", 40, 59);
      g.drawString("Data packet", 40, 79);
     }
    if (fixed.algo == 1)     // draws the key for flooding.  
     {
      g.setColor(Color.lightGray);
      g.fill3DRect(5, 50, 170, 30, true);
      g.setColor(Color.green);
      g.drawOval(13, 64, 3, 3);  
      g.fillOval(13, 64, 3, 3);  
      g.setColor(Color.black);
      g.drawString("  -  Duplicate packets discarded.", 18, 68);
     } 
    for (i = 0; i < 8; i++)        // draw the network on canvas.
     {
      g.setColor(Color.black);
      x = f.getx(i);     // get position of node 'i'.
      y = f.gety(i);
      g.drawString(String.valueOf(i + 1), x + 5, y - 5);  // print node no. for node 'i'.
      for (j = 0; j < 8; j++)    // draw all the links from node 'i'.
       {
        if (f.getDist(i, j) != 0) 
         {
          g.setColor(Color.black);
          g.drawLine(x + 5, y + 5, f.getx(j) + 5, f.gety(j) + 5);  // draw the link.
          xtmp = (x + f.getx(j)) / 2;
          ytmp = (y + f.gety(j)) / 2;
          g.setColor(Color.yellow);
          g.drawString(String.valueOf(f.getDist(i, j)), xtmp, ytmp);  // print the weight of the link.
         } 
       }
     }
    for (i = 0; i < 8; i++)    // draws the nodes.
     {
      x = f.getx(i);
      y = f.gety(i);
      if (redorblue[i] == true) g.setColor(Color.red);
       else g.setColor(Color.blue);
      g.drawOval(x, y, 10, 10);
      g.fillOval(x, y, 10, 10);
     }
    if (fixed.algo == 0) 
     if ((packet.xpos != 0) && (packet.ypos != 0))  // for redrawing the packet in case it has been erased by an open window.
      {
       g.setColor(Color.darkGray);
       g.fill3DRect(packet.xpos, packet.ypos, packet.length, packet.width, true);
      }
      else 
       {
        g.setColor(Color.lightGray);
        g.fillRect(packet.xpos - 1, packet.ypos - 1, packet.length + 2, packet.width + 2);
       }
   }
   
  public void activateNodes(Graphics g, int a, int b) // paints source and destination nodes in red.
   {
    g.setColor(Color.red);
    g.drawOval(f.getx(a), f.gety(a), 10, 10);
    g.fillOval(f.getx(a), f.gety(a), 10, 10);
    g.drawOval(f.getx(b), f.gety(b), 10, 10);
    g.fillOval(f.getx(b), f.gety(b), 10, 10);
    redorblue[a] = redorblue[b] = true;
   }
  
  public void sendInfo(int a, int b, int r[][], int d)  // handles the simulation.
   {
    int i, j;
    Graphics g = getGraphics();
    if (fixed.algo == 0)           // fixed routing.
     {
      packet.sendPacket(g, a, b, r, d);     // this completes the entire simulation.
      redorblue[a] = redorblue[b] = false;  // reset.
      repaint();
     } 
     else if (fixed.algo == 1)     // flooding.
      {
       Packet.first = a;           // source node.
       Packet.last = b;            // destination node.
       packets = -1;             
       for (i = 0; i < 8; i++)     // loop for sending packets from source node.
        {
         if (f.getDist(a, i) != 0)   // there should be a link between 'a' and 'i'. 
          {
           packets++;          
           fpacket[packets] = new Packet(a, a, i, d);   // create a packet
           threads[packets] = new Thread(this);         // and thread for
           threads[packets].start();                    // packet to be sent.
           try                             
            {
             Thread.sleep(100);
            }
            catch(InterruptedException e) {}
          } 
        } 
       for (i = 0; i < 8; i++)    // source node 'i' has sent packets
        Packet.sent[i][a] = true; // so anymore reaching it will be discarded.    
       while (true)               // wait for simulation to end.
        {
         i = packets;
         try
          {
           threads[i].join();
          }
          catch(InterruptedException e) {}
         try
          {
           Thread.sleep(40 * d);
          }
          catch(InterruptedException e) {}
         if (i == packets) break;
        }
       for (i = 0; i < 8; i++)           // reset.
        for (j = 0; j < 8; j++)
         Packet.sent[i][j] = false;
       redorblue[a] = redorblue[b] = false; 
       repaint();
      }  
   }
   
  public void run()    // thread for each packet that is sent(flooding).
   {
    int j, k, temp = packets;  // store value of packets in temp to avoid thread synchronisation problems.
    Graphics g = getGraphics();
    
/* Three checks to be performed before packet is sent(the first 'if' statement) :
    1)Has packet already been sent from this node? If yes, discard packet.
    2)Do not send packet to node from which it has been received.
    3)Is this the destination node? If yes, do not send packet anywhere. */
   
    if ((Packet.sent[fpacket[temp].from][fpacket[temp].source] == false) && (fpacket[temp].dest != fpacket[temp].from) && (fpacket[temp].source != Packet.last))
     {
      fpacket[temp].sendPacket(g, colors[fpacket[temp].source], temp);  // sends packet to next node.
      for (j = 0; j < 8; j++)   // from destination node prepare to send packets to other nodes(j).
       {
        if (f.getDist(fpacket[temp].dest, j) != 0)
         {
          packets++;
          fpacket[packets] = new Packet(fpacket[temp].source, fpacket[temp].dest, j, Packet.delay);
          threads[packets] = new Thread(this);
          threads[packets].start();
          try                             
           {
            Thread.sleep(100);
           }
          catch(InterruptedException e) {}
         }
       }
      Packet.sent[fpacket[temp].from][fpacket[temp].source] = true;  // indicates that this node has already sent packets so subsequent ones are to be discarded.
     }
     else if ((Packet.sent[fpacket[temp].from][fpacket[temp].source] == true))  // if packets have already been sent from this node, the incoming packet is discarded.
      {
       g.setColor(Color.lightGray);
       g.fillRect(fpacket[temp].xpos - 1, fpacket[temp].ypos - 1, Packet.length + 2, Packet.width + 2);   // clear previous position of packet.
       g.setColor(Color.green);
       g.drawOval(fpacket[temp].xpos + 3, fpacket[temp].ypos - 2, 3, 3);  // draw indicator(duplicate packet discarded).
       g.fillOval(fpacket[temp].xpos + 3, fpacket[temp].ypos - 2, 3, 3);  
      }
    threads[temp].stop();   // stop thread once packet has reached next node.
   }
 }  
 
class Window extends Frame   // window for displaying routing table.
 {
  Canvas2 canvas;
  ScrollPane scrollpane;
   
  Window(String title)
   {
    super(title);
    setLayout(new BorderLayout());
    enableEvents(AWTEvent.WINDOW_EVENT_MASK);
    scrollpane = new ScrollPane();
    add("Center", scrollpane);
    canvas = new Canvas2();
    canvas.setSize(300, 300);
    scrollpane.add(canvas);
   }
    
  public void processWindowEvent(WindowEvent event)
   {
    if (event.getID() == WindowEvent.WINDOW_CLOSING) setVisible(false);
   }
 }

class Canvas2 extends Canvas     // for drawing the routing table.
 {
  Font font;  
  int rtable[][] = new int[8][8];

  public void paint(Graphics g)
   {
    int i, j, temp;
    setBackground(Color.lightGray);
    g.setColor(Color.red);
    g.setFont(font = new Font("Times New Roman", Font.PLAIN |  Font.BOLD, 10));
    g.drawString("From", 5, 70);
    g.drawString("Node", 5, 80);
    g.drawString("To Node", 50, 40);
    g.setFont(font = new Font("Times New Roman", Font.PLAIN, 12));
    g.setColor(Color.black);
    g.drawRect(50, 60, 240, 200);
    for (i = 0; i < 8; i++)
     {
      g.setColor(Color.black);
      g.drawLine(50, (i * 25) + 85, 290, (i * 25) + 85);   // make grid.
      g.drawLine((i * 30) + 80, 60, (i * 30) + 80, 260);
      g.drawString(String.valueOf(i + 1), 64 + (i * 30), 55);
      g.drawString(String.valueOf(i + 1), 40, 76 + (i * 25));
      g.setColor(Color.yellow);
      for (j = 0; j < 8; j++)       // fill in the values.
       {
        temp = rtable[i][j] + 1;
        if (temp == 9) g.drawString("-", 63 + (j * 30), 76 + (i * 25));  // no path exists between this pair of nodes.
         else g.drawString(String.valueOf(temp), 63 + (j * 30), 76 + (i * 25));   // path exists.
       }
     } 
   }
    
  public void copyRtable(int temp[][])   // gets routing table entries from 'fixed' class.
   {
    int i, j;
    for (i = 0; i < 8; i++)
     for (j = 0; j < 8; j++)
      rtable[i][j] = temp[i][j];
   }
 }
 
class Win extends Frame implements ActionListener, TextListener    // window for entering source and destination nodes.
 {
  int source, dest; 
  boolean done = false;
  Label fill; 
  Label snodel;
  Label dnodel;
  TextField snode;
  TextField dnode;
  Button ok;
  Button cancel;
  String stmp = new String();
  String dtmp = new String();
  
  Win(String title)
   {
    super(title);
    enableEvents(AWTEvent.WINDOW_EVENT_MASK);
    setBackground(Color.lightGray);
    GridBagLayout gb = new GridBagLayout();
    GridBagConstraints gbc = new GridBagConstraints();
    setLayout(gb);
    gbc.gridx = 0;
    gbc.gridy = 0;
    gbc.gridheight = 1;
    gbc.gridwidth = 1;
    fill = new Label();
    snode = new TextField(1);
    dnode = new TextField(1);
    snodel = new Label("Packet starts from node   : ");
    dnodel = new Label("Packet has to reach node : ");
    ok = new Button("  Ok  ");
    cancel = new Button("Cancel");

    gbc.gridwidth = 2;
    add(snodel, gbc);

    gbc.gridx = 2;
    gbc.gridwidth = 1;
    add(snode, gbc);

    gbc.gridy++;
    gbc.gridx = 0;
    gbc.gridwidth = 2;
    add(dnodel, gbc);

    gbc.gridx = 2;
    gbc.gridwidth = 1;
    add(dnode, gbc);

    gbc.gridy++;
    add(fill, gbc);

    gbc.gridy++;
    gbc.gridx = 0;
    add(ok, gbc);

    gbc.gridx = 2;
    add(cancel, gbc);

    ok.addActionListener(this);
    cancel.addActionListener(this);
    snode.addTextListener(this);
    dnode.addTextListener(this);
    ok.setEnabled(false);
   }

  public void processWindowEvent(WindowEvent event)
   {
    if (event.getID() == WindowEvent.WINDOW_CLOSING)
     {
      snode.setText("");      // reset source and destination values before closing window.
      dnode.setText("");
      setVisible(false);
     }
   }
   
  public void actionPerformed(ActionEvent event)
   {
    if (event.getSource() == ok) // OK clicked - source and destination values obtained from user.
     {
      source=Integer.parseInt(snode.getText());
      dest=Integer.parseInt(dnode.getText());
      if((source>0) && (source<9) && (dest>0) && (dest<9))
        done=true;
      else
        done=false;

      snode.setText("");     // reset values of source and destination.
      dnode.setText("");
      setVisible(false);
      done = true;
     } 
    if (event.getSource() == cancel)    // Cancel clicked.
     {
      snode.setText("");
      dnode.setText("");
      setVisible(false);
     } 
   }   
   
  public void textValueChanged(TextEvent event)
   {
    String temp;
    temp = snode.getText();      // takes care that only single digit is entered for source.
    if (temp.length() > 1) snode.setText(stmp);
    stmp = snode.getText(); 
    
    temp = dnode.getText();      // takes care that only single digit is entered for destination.
    if (temp.length() > 1) dnode.setText(dtmp);
    dtmp = dnode.getText(); 
    
    if (done == false)
     {
      if (event.getSource() == snode)  // get value of source.
       {
        try 
         {
          source = Integer.parseInt(snode.getText());
         }
         catch(NumberFormatException e)
          {
           source = 0;
          }
       } 
      if (event.getSource() == dnode)   // get value of destination.
       {
        try
         {
          dest = Integer.parseInt(dnode.getText());
         }
         catch(NumberFormatException e)
          {
           dest = 0;
          }
       } 
     } 
    if ((source > 0) && (source < 9) && (dest > 0) && (dest < 9))  // check for valid values of source and destination.
      ok.setEnabled(true);
     else ok.setEnabled(false);
   }
   
  public int getSource()    // give source value to 'fixed' class.
   {
    return(source - 1);
   }
   
  public int getDest()
   {
    return(dest - 1);
   }
   
  public boolean getDone()
   {
    return(done);
   }
 }
 
class Options extends Frame implements ActionListener   // window for options.
 {
  int delay = 50, delay1, algo1;      // speed is medium by default(delay = 50).
  Menu simspeed, changes, algorithm;
  MenuBar menubar;
  MenuItem m11, m12, m13;    // speed.
  MenuItem m31, m32;         // algorithm.
  MenuItem m21;              // changes.
  
  Options(String title)
   {
    super(title);
    enableEvents(AWTEvent.WINDOW_EVENT_MASK);
    setBackground(Color.lightGray);
    setLayout(new FlowLayout());
    menubar = new MenuBar();
    simspeed = new Menu("Speed of Simulation");
    algorithm = new Menu("Routing Algorithm");
    changes = new Menu("Changes");

    m11 = new MenuItem("Slow");
    simspeed.add(m11);
    m11.addActionListener(this);

    m12 = new MenuItem("Medium");
    simspeed.add(m12);
    m12.addActionListener(this);

    m13 = new MenuItem("Fast");
    simspeed.add(m13);
    m13.addActionListener(this);
    
    m31 = new MenuItem("Fixed Routing");
    algorithm.add(m31);
    m31.addActionListener(this);
    
    m32 = new MenuItem("Flooding");
    algorithm.add(m32);
    m32.addActionListener(this);
    
    m21 = new MenuItem("Save changes");
    changes.add(m21);
    m21.addActionListener(this);

    menubar.add(simspeed);
    menubar.add(algorithm);
    menubar.add(changes);
    setMenuBar(menubar);
   } 

  public void processWindowEvent(WindowEvent event)
   {
    if (event.getID() == WindowEvent.WINDOW_CLOSING)   // exit window without saving changes.
     {
      setSize(500, 150);
      setVisible(false);
     }
   }
   
  public void actionPerformed(ActionEvent event)
   {
    if (event.getSource() == m11) delay1 = 500;   // slow speed.
    if (event.getSource() == m12) delay1 = 50;    // medium speed.
    if (event.getSource() == m13) delay1 = 5;     // fast speed.
    if (event.getSource() == m31) algo1 = 0;      // fixed routing.
    if (event.getSource() == m32) algo1 = 1;      // flooding.
    if (event.getSource() == m21)                 // save changes.
     {
      delay = delay1;          // if changes are to be saved,
      fixed.algo = algo1;      // save them and close window.
      if (fixed.algo == 0)
       {
        fixed.start.setEnabled(true);
        fixed.table.setEnabled(true);
       }
      if (fixed.algo == 1) 
       {
        fixed.start.setEnabled(true);
        fixed.table.setEnabled(false);
       } 
      setSize(500, 150); 
      setVisible(false);
     } 
   }
 }
 
class Packet extends Applet     // class for handling the data packet.
 {
  fixed f = new fixed();
  static int length = 16;       // dimensions of packet.
  static int width = 7;      
  int hlength = 6;              // dimensions of packet header.
  int hwidth = 7;                    
  int xpos, ypos;               // position of packet.
  int source, dest;             
  int from;                     // packet has reached current node from 'from' node. 
  static int first, last, delay;

/* 'sent' is required to prevent packet avalanche in flooding.
   If a node has already sent packets, the 'sent' value is made
   true to indicate this. This value is checked before sending
   packet - if True, discard packet.
   2-dimensional array is required to overcome thread synchrony
   problems. Consider the element sent[a][b] - here we are checking
   for node 'b'. 'a' indicates the packet that 'b' will send has 
   come from which node. When 'b' sends a packet received from 'a',
   all the sent[x][b] except for x=a are made true. sent[a][b] is
   made true after all packets have been sent from 'b'.
   sent[a][b] cannot be made true when first packet has been sent
   from 'b' since then the remaining packets from 'b' will not be
   sent(since sent will be true). But now until all packets have been
   sent from 'b', sent is false and any packet coming to 'b' from 
   some other node will also be flooded out from 'b' which will
   end up causing an avalanche.       
*/   
   
  static boolean sent[][] = {{false, false, false, false, false, false, false, false },
                             {false, false, false, false, false, false, false, false },
                             {false, false, false, false, false, false, false, false },
                             {false, false, false, false, false, false, false, false },
                             {false, false, false, false, false, false, false, false },
                             {false, false, false, false, false, false, false, false },
                             {false, false, false, false, false, false, false, false },
                             {false, false, false, false, false, false, false, false }};
  Font font;
  FontMetrics fontm;

  Packet()             // constructor for fixed routing packet.
   {
    xpos = ypos = 0;
   }
   
  Packet(int a, int b, int c, int d)  // constructor for flooding packet.
   {
    from = a;
    source = b;
    dest = c;
    delay = d;
    xpos = f.getx(source);
    ypos = f.gety(source) - 20;
   }
    
  public void sendPacket(Graphics g, int source, int dest, int rtable[][], int delay)    // simulates transfer of packet for Fixed Routing.
   {
    int i = source, j, k, m, x, y, x1, y1, x2, y2, xa, ya;
    g.setFont(font = new Font("Times New Roman", Font.PLAIN, 10));
    g.drawString(String.valueOf(dest + 1), f.getx(source) - 7, f.gety(source) - 21);
    g.setColor(Color.pink);
    g.fill3DRect(f.getx(source) - 7, f.gety(source) - 20, hlength , hwidth, true); 
    g.setColor(Color.darkGray);
    g.fill3DRect(f.getx(source), f.gety(source) - 20, length, width, true);
    try                             // startup delay.
     {
      Thread.sleep(20 * delay);
     }
     catch(InterruptedException e) {}
    while (i != dest)                 // checks if destination has been reached.
     {
      j = rtable[i][dest];            // 'j' is next node in path.
      xpos = x1 = f.getx(i);          // initialise position of packet.
      ypos = y1 = f.gety(i) - 20;
      x2 = f.getx(j);                 // position of destination node.
      y2 = f.gety(j) - 20;
      xa = (x2 - x1) / 10;            // distance packet moves by each time.
      ya = (y2 - y1) / 10;

      while (xpos != x2)              // while next node in path has not been reached.
       {
        g.setColor(Color.lightGray);
        g.fillRect(xpos - 8, ypos - 10, length + 10, width + 19);   // clear previous position of packet.
        xpos += xa;       // change position of packet.
        ypos += ya;
        if (((xpos > x2) && (xa > 0)) || ((xpos < x2) && (xa < 0)))  // check if packet has passed destination node.
         xpos = x2;
        if (((ypos > y2) && (ya > 0)) || ((ypos < y2) && (ya < 0)))
         ypos = y2;
        g.setColor(Color.black); 
        g.drawString(String.valueOf(dest + 1), xpos - 7, ypos - 1);
        g.setColor(Color.pink);
        g.fill3DRect(xpos - 7, ypos, hlength , hwidth, true); 
        g.setColor(Color.darkGray);
        g.fill3DRect(xpos, ypos, length, width, true); // draw packet at new position. 

        for (k = 0; k < 8; k++)   // refresh display since packet erases some part of network.
         {
          x = f.getx(k);
          y = f.gety(k);
          g.drawString(String.valueOf(k + 1), x + 5, y - 5);
          for (m = 0; m < 8; m++)
           {
            if (f.getDist(k, m) != 0) 
             {
              g.setColor(Color.yellow);
              g.drawString(String.valueOf(f.getDist(k, m)), (x + f.getx(m)) / 2, (y + f.gety(m)) / 2);
              g.setColor(Color.black);
              g.drawLine(x + 5, y + 5, f.getx(m) + 5, f.gety(m) + 5);
             } 
           }
         }
        for (k = 0; k < 8; k++)            // refreshing display.
         {
          x = f.getx(k);
          y = f.gety(k);
          if ((k == source) || (k == dest)) g.setColor(Color.red);
           else g.setColor(Color.blue);
          g.drawOval(x, y, 10, 10);
          g.fillOval(x, y, 10, 10);
         }

        try                               // delay to regulate speed of simulation.
         {
          Thread.sleep(delay * f.getDist(i, j));
         }
         catch(InterruptedException e) {}
       } 
      i = j;                        // destination is now source for next iteration.      
      try                           // delay at each node. 
       {
        Thread.sleep(2 * delay);
       }
       catch(InterruptedException e) {}
     }
    xpos = ypos = 0;                       // resetting position of packet.
   }
   
  public void sendPacket(Graphics g, Color color, int temp)    // simulates transfer of packet for flooding.
   {
    int k, m, x, y, x1, y1, x2, y2, xa, ya;
    for (k = 0; k < 8; k++)                   // node 'source' has sent a packet - prevents avalanche.
     if (k != from) Packet.sent[k][source] = true;
    g.setFont(font = new Font("Times New Roman", Font.PLAIN, 10));
    g.setColor(color);
    g.fill3DRect(f.getx(source), f.gety(source) - 20, length, width, true);
    try                             // startup delay.
     {
      Thread.sleep(20 * delay);
     }
     catch(InterruptedException e) {}
    x1 = xpos;                      // initialise position of packet.
    y1 = ypos;
    x2 = f.getx(dest);              // position of destination node.
    y2 = f.gety(dest) - 20;
    xa = (x2 - x1) / 10;            // distance packet moves by each time.
    ya = (y2 - y1) / 10;

    while (xpos != x2)              // while destination has not been reached.
     {
      g.setColor(Color.lightGray);
      g.fillRect(xpos - 1, ypos - 1, length + 2, width + 2);   // clear previous position of packet.
      xpos += xa;       // change position of packet.
      ypos += ya;
      if (((xpos > x2) && (xa > 0)) || ((xpos < x2) && (xa < 0)))  // check if packet has passed destination node.
       xpos = x2;
      if (((ypos > y2) && (ya > 0)) || ((ypos < y2) && (ya < 0)))
       ypos = y2;
      g.setColor(color);
      g.fill3DRect(xpos, ypos, length, width, true); // draw packet at new position. 
      g.setColor(Color.black);

      for (k = 0; k < 8; k++)   // refresh display since packet erases some part of network.
       {
        x = f.getx(k);
        y = f.gety(k);
        g.drawString(String.valueOf(k + 1), x + 5, y - 5);
        for (m = 0; m < 8; m++)
         {
          if (f.getDist(k, m) != 0) 
           {
            g.setColor(Color.yellow);
            g.drawString(String.valueOf(f.getDist(k, m)), (x + f.getx(m)) / 2, (y + f.gety(m)) / 2);
            g.setColor(Color.black);
            g.drawLine(x + 5, y + 5, f.getx(m) + 5, f.gety(m) + 5);
           } 
         }
       }
      for (k = 0; k < 8; k++)            // refreshing display.
       {
        x = f.getx(k);
        y = f.gety(k);
        if ((k == first) || (k == last)) g.setColor(Color.red);
         else g.setColor(Color.blue);
        g.drawOval(x, y, 10, 10);
        g.fillOval(x, y, 10, 10);
       }

      try                           // delay to regulate speed of simulation.
       {
        Thread.sleep(delay * f.getDist(source, dest));
       }
       catch(InterruptedException e) {}
     } 
   }
 }
	
	
class Help extends Frame    // window for displaying the help.
 {
	ScrollPane scrollpane = new ScrollPane(ScrollPane.SCROLLBARS_AS_NEEDED);
        String str ="\nIntroduction\n"
	             + "=========\n\n" 
					     + "This applet simulates the sliding window protocol. It works for the 1-bit, 2-bit and 3-bit protocols.\n"
					     + "By default the 3-bit protocol is used for the ideal case. The simulation can also be run for cases\n"
						+ "where frame losses occur. \n\n\n"
					     + "Changing the Settings\n"
	             + "==================\n\n" 
					     + "Changing the settings for the simulation is quite straightforward. Click on the 'Change Settings'\n"
					     + "button. The caption changes to 'Done' and a window comes up which offers you a number of options.\n"
					     + "There are textboxes in which you can type in the values and there are also drop-down lists from\n"
					     + "which you can select the desired option just by clicking on it. The options related to losses are\n"
					     + "available only when a loss option is selected from the 'Type of Simulation' field. When you have\n"
					     + "finished with the settings, you can exit by clicking on the 'Done' button whose caption now reverts\n"
							 + "to the original and the options window closes.\n\n\n"	 
					     + "Managing the Simulation\n"
	             + "====================\n\n" 
					     + "To start the simulation, click on the 'Start' button. This causes the 'Stop' and 'Pause' buttons to\n"
					     + "get enabled. As the name suggests, the 'Stop' button, when clicked, stops the simulation and\n"
					     + "restores the screen to the state it was in before the simulation started. The simulation can be\n"
					     + "restarted in the same way as before and the settings can be changed if required. Clicking on the\n"
					     + "'Pause' button during the simulation freezes the simulation. It can be resumed by clicking on the\n"
							 + "'Resume' button.";	
	TextArea text = new TextArea(str, 40, 77, TextArea.SCROLLBARS_NONE);
  
	Help(String title)
	 {
		super(title);
    setLayout(new BorderLayout());
    enableEvents(AWTEvent.WINDOW_EVENT_MASK);
    scrollpane = new ScrollPane();
    add("Center", scrollpane);
    scrollpane.add(text);
		text.setEnabled(false);
	 }

  public void processWindowEvent(WindowEvent event)
   {
    if (event.getID() == WindowEvent.WINDOW_CLOSING) setVisible(false);
   }
 }	


//<applet code = fixed.class width = 620 height = 430></applet>
